#include <stdio.h>
#include <stdlib.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <errno.h>
#include <error.h>

#include <linux/input.h>

#include <xorg/xorg-server.h>
#include <xorg/xf86.h>

extern const char *__progname;

/* Taken from synaptics-0.14.6/eventcomm.c */

#define SYSCALL(call) while (((call) == -1) && (errno == EINTR))

#define LONG_BITS (sizeof(long) * 8)
#define NBITS(x) (((x) + LONG_BITS - 1) / LONG_BITS)
#define OFF(x)   ((x) % LONG_BITS)
#define LONG(x)  ((x) / LONG_BITS)
#define TEST_BIT(bit, array) (array[LONG(bit)] & (1 << OFF(bit)))

static Bool
event_query_is_touchpad(int fd)
{
    int ret;
    unsigned long evbits[NBITS(KEY_MAX)];

    /* Check for ABS_X, ABS_Y, ABS_PRESSURE and BTN_TOOL_FINGER */

    SYSCALL(ret = ioctl(fd, EVIOCGBIT(0, EV_MAX), evbits));
    if (ret < 0)
	return FALSE;
    if (!TEST_BIT(EV_SYN, evbits) ||
	!TEST_BIT(EV_ABS, evbits) ||
	!TEST_BIT(EV_KEY, evbits))
	return FALSE;

    SYSCALL(ret = ioctl(fd, EVIOCGBIT(EV_ABS, KEY_MAX), evbits));
    if (ret < 0)
	return FALSE;
    if (!TEST_BIT(ABS_X, evbits) ||
	!TEST_BIT(ABS_Y, evbits) ||
	!TEST_BIT(ABS_PRESSURE, evbits))
	return FALSE;

    SYSCALL(ret = ioctl(fd, EVIOCGBIT(EV_KEY, KEY_MAX), evbits));
    if (ret < 0)
	return FALSE;
    if (!TEST_BIT(BTN_TOOL_FINGER, evbits))
	return FALSE;

    return TRUE;
}

int
main(int argc, char **argv) {
	int i = 1;
	int rc = EXIT_FAILURE;

	if (argc == 1) {
		fprintf(stderr, "Usage: %s <devname>[<devname>...]\n", __progname);
		return EXIT_FAILURE;
	}
	
	while (i < argc) {
		struct stat st;
		int fd;
	
		if (stat(argv[i], &st) == -1)
			error(EXIT_FAILURE, errno, "stat: %s", argv[i]);

		if (!S_ISBLK(st.st_mode) && !S_ISCHR(st.st_mode)) {
			i++;
			continue;
		}

		if ((fd = open(argv[i], O_RDONLY)) == -1)
			error(EXIT_FAILURE, errno, "%s: open", argv[i]);

		if (event_query_is_touchpad(fd)) {
			fprintf(stdout, "%s\n", argv[i]);
			rc = EXIT_SUCCESS;
		}

		if (close(fd) == -1)
			error(EXIT_FAILURE, errno, "%s: close", argv[i]);
		i++;
	}
	
	return rc;
}
