/* -*- mode: C; tab-width: 8; fill-column: 70; -*-  */

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <getopt.h>
#include <errno.h>
#include <error.h>

#include <xf86Parser.h>
#include <xf86Optrec.h>
#include <X11/Intrinsic.h>

#define CONFPATH "%A,%R,/etc/X11/%R,%P/etc/X11/%R,%E,%F,/etc/X11/%F,"	\
	"%P/etc/X11/%F,%D/%X,/etc/X11/%X,/etc/%X,%P/etc/X11/%X.%H,"	\
	"%P/etc/X11/%X,%P/lib/X11/%X.%H,%P/lib/X11/%X"

#define PACKAGE_VERSION "0.0.1"

#define ADD 1
#define DEL 2

extern const char *__progname;
static int action = -1;
static int show_section_name = 0;

static int
add_to_layout (XF86ConfigPtr config, XF86ConfInputPtr input, char *type) {
	XF86ConfLayoutPtr layout = config->conf_layout_lst;
	XF86ConfInputrefPtr inputref = NULL, new_inputref;

	if ((new_inputref = (XF86ConfInputrefRec *) calloc (1,sizeof(XF86ConfInputrefRec))) == NULL)
		error (EXIT_FAILURE, errno, "calloc");

	new_inputref->list.next = NULL;
	new_inputref->iref_inputdev = input;
	new_inputref->iref_inputdev_str = XtNewString (input->inp_identifier);
	new_inputref->iref_option_lst = xf86addNewOption (new_inputref->iref_option_lst,
							  XtNewString (type),
							  NULL);

	for (inputref = layout->lay_input_lst;
	     layout->lay_input_lst && inputref->list.next;
	     inputref = inputref->list.next) { }

	if (!inputref)
		layout->lay_input_lst = new_inputref;
	else
		inputref->list.next = new_inputref;

	return EXIT_SUCCESS;
}

static int
is_mouse (XF86ConfInputrefPtr inputref) {
	XF86OptionPtr p;
	for (p = inputref->iref_option_lst; p; p = xf86nextOption (p)) {
		if (!strcmp (xf86optionName (p), "CorePointer") ||
		    !strcmp (xf86optionName (p), "AlwaysCore")  ||
		    !strcmp (xf86optionName (p), "SendCoreEvents"))
			return 1;
	}
	return 0;
}

static int
remove_input (XF86ConfigPtr config, char *identifier) {
	XF86ConfLayoutPtr layout = config->conf_layout_lst;
	XF86ConfInputrefPtr inputref, prev_inputref, next_inputref;
	XF86ConfInputPtr input, prev_input;

	if (!identifier)
		error (EXIT_FAILURE, 0, "Identifier not specified");

	if (!layout->lay_input_lst)
		error (EXIT_FAILURE, 0, "InputDevice not found in ServerLayout section");

	inputref = layout->lay_input_lst;
	prev_inputref = NULL;

	while (inputref) {
		if (!strcmp (identifier, inputref->iref_inputdev_str)) {
			if (!prev_inputref)
				layout->lay_input_lst = inputref->list.next;
			else
				prev_inputref->list.next = inputref->list.next;

			xf86optionListFree (inputref->iref_option_lst);
			free (inputref->iref_inputdev_str);
			free (inputref);
			break;
		}

		prev_inputref = inputref;
		inputref = inputref->list.next;
	}

	if (layout->lay_input_lst) {
		int found_core_pointer = 0;

		for (inputref = layout->lay_input_lst; inputref; inputref = inputref->list.next) {
			XF86OptionPtr p;

			for (p = inputref->iref_option_lst; p; p = xf86nextOption (p))
				if (!strcmp (xf86optionName (p), "CorePointer")) {
					found_core_pointer = 1;
					break;
				}
		}

		inputref = layout->lay_input_lst;

		if (!found_core_pointer && is_mouse (inputref)) {
			xf86optionListFree (inputref->iref_option_lst);
			inputref->iref_option_lst = xf86newOption ("CorePointer", NULL);
		}
	}

	input = config->conf_input_lst;
	prev_input = NULL;

	while (input) {
		if (!strcmp (identifier, input->inp_identifier)) {
			if (!prev_input)
				config->conf_input_lst = input->list.next;
			else
				prev_input->list.next = input->list.next;

			xf86optionListFree (input->inp_option_lst);
			free (input->inp_identifier);
			free (input->inp_driver);
			free (input);
			break;
		}

		prev_input = input;
		input = input->list.next;
	}

	return EXIT_SUCCESS;
}

static char *
extract_value (char *line) {
	char *value = NULL, *e = NULL;

	value = strchr (line, '"');

	if (!value) return NULL;
	value++;

	e = strchr (value, '"');
	if (!e) return NULL;
	*e = '\0';
	
	return value;
}

static int
add_input (XF86ConfigPtr config) {
	XF86ConfInputPtr new_input, last = NULL, next = NULL;
	char *line = NULL;
	size_t len = 0;

	if ((new_input = (XF86ConfInputRec *) calloc (1,sizeof(XF86ConfInputRec))) == NULL)
		error (EXIT_FAILURE, errno, "calloc");

	new_input->list.next = NULL;
	new_input->inp_option_lst = NULL;

	while (getline (&line, &len, stdin) != -1) {
		char *value = NULL;

		if (!strncasecmp (line, "Identifier", 10)) {
			value = (line+10);
			
			if ((value = extract_value (value)) == NULL)
				continue;

			new_input->inp_identifier = XtNewString (value);
		}
		else if (!strncasecmp (line, "Driver", 6)) {
			value = (line+6);

			if ((value = extract_value (value)) == NULL)
				continue;

			new_input->inp_driver = XtNewString (value);
		}
		else if (!strncasecmp (line, "Option", 6)) {
			char *name = (line+6);
			size_t nlen;

			if ((name = extract_value (name)) == NULL)
				continue;

			if ((nlen = strlen (name)) >= len)
				continue;
			
			if ((value = extract_value (name+nlen+1)) == NULL)
				continue;

			new_input->inp_option_lst = xf86addNewOption (new_input->inp_option_lst,
								      XtNewString (name),
								      XtNewString (value));
		}
	}

	if (!new_input->inp_identifier || !new_input->inp_driver)
		error (EXIT_FAILURE, 0, "You should specify Identifier and Driver");

	if (config->conf_layout_lst) {
		XF86ConfLayoutPtr layout = config->conf_layout_lst;
		XF86ConfInputrefPtr inputref = NULL, last_inputref = NULL;
		
		for (inputref = layout->lay_input_lst; inputref; inputref = inputref->list.next) {
			if (!is_mouse (inputref))
				continue;

			if (!strcmp (new_input->inp_identifier, inputref->iref_inputdev_str))
				error (EXIT_FAILURE, 0, "Identifier already exist!");

			if (inputref->list.next == NULL)
				last_inputref = inputref;
		}
		last_inputref ?
			add_to_layout (config, new_input, "AlwaysCore") :
			add_to_layout (config, new_input, "CorePointer");
	}


	for (next = config->conf_input_lst; next; next = next->list.next) {
		if (!strcmp (new_input->inp_identifier, next->inp_identifier))
			error (EXIT_FAILURE, 0, "Identifier already exist!");

		if (next->list.next == NULL)
			last = next;
	}

	if (!last)
		config->conf_input_lst = new_input;
	else
		last->list.next = new_input;

	return EXIT_SUCCESS;
}

static void
show_input_section (XF86ConfInputPtr s) {
	XF86OptionPtr opt;
	if (show_section_name) {
		printf ("%s\n", s->inp_identifier);
		return;
	}
	printf ("Section \"InputDevice\"\n");
	printf ("\tIdentifier \"%s\"\n", s->inp_identifier);
	printf ("\tDriver \"%s\"\n", s->inp_driver);
	xf86printOptionList (stdout, s->inp_option_lst, 1);
	printf ("EndSection\n");
}

static int
getinfo (XF86ConfigPtr config, char *name) {
	int rc = EXIT_FAILURE;

	XF86ConfLayoutPtr layout = config->conf_layout_lst;
	XF86ConfInputrefPtr inputref = layout->lay_input_lst;

	for (inputref = layout->lay_input_lst; inputref; inputref = inputref->list.next) {
		XF86OptionPtr p;

		for (p = inputref->iref_option_lst; p; p = xf86nextOption (p)) {
			if (strcmp (xf86optionName (p), "CorePointer") &&
			    strcmp (xf86optionName (p), "AlwaysCore")  &&
			    strcmp (xf86optionName (p), "SendCoreEvents"))
				    continue;

			if (name) {
				if (!strcmp (inputref->iref_inputdev_str, name)) {
					show_input_section (inputref->iref_inputdev);
					rc = EXIT_SUCCESS;
				}
			}
			else {
				show_input_section (inputref->iref_inputdev);
				rc = EXIT_SUCCESS;
			}
		}

	}
	return rc;
}

static struct option long_options[] = {
	{"help", 0, 0, 'h'},
	{"version", 0, 0, 'V'},
	{"verbose", 0, 0, 'v'},
	{"add", 0, 0, 'a'},
	{"del", 0, 0, 'd'},
	{"sections", 0, 0, 's'},
	{"config", 1, 0, 'c'},
	{0, 0, 0, 0}
};

static void __attribute__ ((noreturn))
print_help (int ret) {
	printf ("Usage: %s [OPTIONS] [<Identifier>]\n", __progname);
	printf ("   or: %s [OPTIONS] -d <Identifier>\n", __progname);
	printf ("   or: %s [OPTIONS] -a\n", __progname);
	printf ("Options:\n");
	printf ("  -a, --add            Add new input section.\n");
	printf ("  -d, --del            Remove input section.\n");
	printf ("  -c, --config=CONFIG  Specify Xorg config file.\n");
	printf ("  -s, --sections       Show only section names.\n");
	printf ("  -h, --help           Output a brief help message.\n");
	printf ("  -V, --version        Print program version and exit.\n");
	printf ("\n");
	exit (ret);
}

static void __attribute__ ((noreturn))
print_version (void) {
        printf ("%s version %s\n"
               "\nCopyright (C) 2007  Alexey Gladkov <legion@altlinux.org>\n"
               "\nThis is free software; see the source for copying conditions.\n"
               "There is NO warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n"
               "\nWritten by Alexey Gladkov <legion@altlinux.org>\n",
               __progname, PACKAGE_VERSION);
        exit (EXIT_SUCCESS);
}

int
main (int argc, char **argv) {
	int c, rc = EXIT_SUCCESS;
	const char *filename;
	char *xconf = NULL, *opt_name = NULL, *opt_type = NULL;
	XF86ConfigPtr config;

	while (1) {
		int option_index = 0;
		c = getopt_long (argc, argv, "adc:shV", long_options, &option_index);

		if (c == -1)
			break;

		switch (c) {
		case 'a':
			action = ADD;
			break;
		case 'c':
			xconf = optarg;
			break;
		case 'd':
			action = DEL;
			break;
		case 's':
			show_section_name = 1;
			break;
		case 'V':
			print_version ();
			break;
		default:
		case 'h':
			print_help (EXIT_SUCCESS);
			break;
		}
	}
	
	if (argc > optind)
		opt_name = argv[optind++];

	if (!xconf)
		xconf = "./xorg.conf";

	if (!(filename = xf86openConfigFile (CONFPATH, xconf, NULL)))
		error (EXIT_FAILURE, 0, "Unable to open config file");

	if ((config = xf86readConfigFile ()) == NULL)
		error (EXIT_FAILURE, 0, "Problem when parsing config file");

	xf86closeConfigFile ();

	switch (action) {
	case ADD:
		rc = add_input (config);
		break;
	case DEL:
		rc = remove_input (config, opt_name);
		break;
	default:
		rc = getinfo (config, opt_name);
	}

	if (rc == EXIT_SUCCESS && (action == ADD || action == DEL)) {
//		fprintf (stderr, "DEBUG: Write\n");
		if (!(xf86writeConfigFile (xconf, config)))
			error (EXIT_FAILURE, 0, "Unable to write config file");
	}

	return rc;
}
